(*
 * Constants and types used by the build module Omake_build.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2003 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)
open Lm_symbol
open Lm_location

open Omake_ir
open Omake_env
open Omake_node
open Omake_exec
open Omake_cache_type

(*
 * Magic number for marshaling.
 *)
let magic_number = Omake_magic.env_magic

(*
 * A command corresponds to a rule,
 * where all variables have been resolved.
 *
 * A command can be in several states:
 *    Idle: the command is not currently used
 *    Initial: the command needs to be processed
 *    ScanBlocked: the command can't be scanned until
 *       all its dependencies exist
 *    Scan: existing dependencies have been satisfied,
 *       and the target should be scanned for additional
 *       dependencies.
 *    Scanning: the command is being scanned, with the given pid
 *    Scanned: the command has been scanned, but at least
 *       one side-effect has not been scanned
 *    Blocked: the command can't be run until
 *       its dependencies are built
 *    Ready: all dependencies have been satisfied, and the
 *       command is ready to run
 *    Pending: command is ready to run, but it conflicts
 *       with another job that is Running
 *    Running pid: the command is running, with the given pid
 *    Finished code: the command exited, with the given code
 *)
type command_state =
   CommandIdle
 | CommandInitial
 | CommandScanBlocked
 | CommandScanReady
 | CommandScanning of Omake_exec_id.id * string
 | CommandScanned
 | CommandBlocked
 | CommandReady
 | CommandPending
 | CommandRunning of Omake_exec_id.id
 | CommandSucceeded
 | CommandFailed of int

type command_tag =
   CommandIdleTag
 | CommandInitialTag
 | CommandScanBlockedTag
 | CommandScanReadyTag
 | CommandScanningTag
 | CommandScannedTag
 | CommandBlockedTag
 | CommandReadyTag
 | CommandPendingTag
 | CommandRunningTag
 | CommandSucceededTag
 | CommandFailedTag

(*
 * A command line can be expanded or not expanded.
 *)
type command_body =
   CommandNone
 | CommandInfo of command_info list
 | CommandLines of venv_command_line list

(*
 * The command.
 * There is a 1-to-1 correspondence between commands and targets.
 *
 *     command_state: the current command state
 *
 *     command_target: the target node to be built
 *     command_effects: the nodes that may be modified by this command
 *     command_loc: the location of the rule being used to build
 *        the target.
 *     command_lines: the commands that should be run sequentially
 *        to build the target.
 *
 *     command_scanner_deps: dependencies that must be
 *        satisfied before the scanner is run.
 *     command_scanner: the name of a command to run that
 *        produces a set of additional dependencies for this file.
 *        The scanner should be run, and dependencies built
 *        before the target is built.
 *     command_scanned_deps: results of running the scanner
 *
 *     command_static_deps: the dependencies statically defined
 *        in the OMakefiles.
 *
 *     command_current_deps: the dependencies, including implicit
 *        dependencies produced by the scanner
 *     command_blocked: at least one of the deps in the set
 *        is blocked
 *)
type command =
   { command_venv                 : venv;
     mutable command_state        : command_state;
     command_target               : Node.t;
     mutable command_effects      : NodeSet.t;
     command_loc                  : loc;

     (* Scanner info *)
     mutable command_scanner_deps : NodeSet.t;
     mutable command_scanner      : command_body;
     mutable command_scanned_deps : NodeSet.t;

     (* Static deps from the OMakefiles *)
     mutable command_static_deps  : NodeSet.t;

     (* Actual dynamic dependencies *)
     mutable command_build_deps   : NodeSet.t;
     mutable command_blocked      : Node.t list;
     mutable command_lines        : command_body;

     (* Linked list *)
     mutable command_pred         : command option ref;
     command_succ                 : command option ref
   }

(*
 * The environment remembers all the commands.  In addition
 * we compute an inverted command graph , so that we know what to do
 * when a command finishes.
 *
 *    env_venv: the default environment
 *    env_cache: the build cache, so we don't run commands
 *       if the target is already up-to-date.
 *    env_explicit_deps: gives explicitly-defined dependencies of nodes
 *    env_explicit_targets: maps nodes to rules where the node is a target
 *    env_explicit_directories: maps directories to environments
 *
 *    env_commands: all the commands
 *    env_inverse: the inverted dependency graph
 *    env_error_code: the exit code when an error occurs
 *    env_idle_count: number of idle processors
 *)
type env =
   { env_venv                       : venv;
     env_cwd                        : Dir.t;
     env_cache                      : Omake_cache.t;
     env_exec                       : exec;
     mutable env_explicit_deps      : NodeSet.t NodeTable.t;
     env_explicit_targets           : erule NodeTable.t;
     env_explicit_directories       : venv DirTable.t;
     env_includes                   : digest NodeTable.t;

     (* Build state *)
     mutable env_commands           : command NodeTable.t;
     mutable env_inverse            : command NodeTable.t NodeTable.t;
     mutable env_error_code         : int;
     mutable env_idle_count         : int;
     mutable env_print_dependencies : NodeSet.t;

     (* The various command queues *)
     env_idle_wl                    : command option ref;
     env_initial_wl                 : command option ref;
     env_scan_blocked_wl            : command option ref;
     env_scan_ready_wl              : command option ref;
     env_scanning_wl                : command option ref;
     env_scanned_wl                 : command option ref;
     env_blocked_wl                 : command option ref;
     env_ready_wl                   : command option ref;
     env_pending_wl                 : command option ref;
     env_running_wl                 : command option ref;
     env_succeeded_wl               : command option ref;
     env_failed_wl                  : command option ref;

     (* Statistics *)
     mutable env_succeeded_count    : int;
     mutable env_scan_count         : int;
     mutable env_scan_exec_count    : int;
     mutable env_rule_count         : int;
     mutable env_rule_exec_count    : int
   }

(*
 * We save only part of the environment to a file.
 *)
type save =
   { save_venv                 : venv;
     save_explicit_deps        : NodeSet.t NodeTable.t;
     save_explicit_targets     : erule NodeTable.t;
     save_explicit_directories : venv DirTable.t;
     save_includes             : digest NodeTable.t
   }


(*
 * Helper type for determining how to build a command
 * from a rule.
 *)
type explicit_rule =
   ExplicitTarget of erule
 | ExplicitDirectory of venv
 | ExplicitNone

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
