(*
 * Utilities on the build environment.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2004 Mojave Group, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)
open Lm_printf
open Lm_symbol

open Omake_env
open Omake_node
open Omake_build_type

module Pos = MakePos (struct let name = "Omake_build_util" end)
open Pos

(*
 * This is a totally different sorting algorithm than that used in
 * revision -r1.2.
 *
 * Here is the new assumption: only direct dependencies matter.
 * That is, the transitive closure is not needed for nodes outside
 * the set being sorted.
 *
 * This version uses a simple DFS to order the nodes.
 *)

(*
 * Get the dependencies for this set of names.
 *)
let command_deps venv orules domain deps =
   let deps = venv_get_ordering_deps venv orules deps in
      NodeSet.fold (fun deps dep ->
            if NodeTable.mem domain dep then
               NodeSet.add deps dep
            else
               deps) NodeSet.empty deps

(*
 * Build the subgraph, including only those nodes that we actually
 * care about.
 *)
let build_subgraph env venv pos orules domain =
   NodeTable.mapi (fun node _ ->
         try
            let command = NodeTable.find env.env_commands node in
            let deps = command_deps venv orules domain command.command_build_deps in
               NodeSet.remove deps node
         with
            Not_found ->
               raise (OmakeException (pos, StringNodeError ("file is not found", node)))) domain

(*
 * Find all the roots.
 *)
let find_roots pos graph =
   let roots =
      NodeTable.fold (fun roots _ deps ->
            NodeSet.fold NodeTable.remove roots deps) graph graph
   in

   (* If the roots are empty, the entire graph is cyclic *)
   let () =
      if NodeTable.is_empty roots then
         let print_graph buf =
            eprintf "@[<hv 3>The entire graph is cyclic!";
            NodeTable.iter (fun node _ -> eprintf "@ %a" pp_print_node node) graph;
            eprintf "@]@."
         in
            raise (OmakeException (pos, LazyError print_graph))
   in
      roots

(*
 * Produce a sort in DFS order.
 *)
let rec dfs_sort_node graph marked items node =
   if NodeSet.mem marked node then
      marked, items
   else
      let marked = NodeSet.add marked node in
      let marked, items = dfs_sort_nodes graph marked items (NodeTable.find graph node) in
         marked, node :: items

and dfs_sort_nodes graph marked items nodes =
   NodeSet.fold (fun (marked, items) node ->
         dfs_sort_node graph marked items node) (marked, items) nodes

let dfs_check pos marked graph =
   if NodeSet.cardinal marked > NodeTable.cardinal graph then
      let nodes = NodeTable.fold (fun marked node _ -> NodeSet.remove marked node) marked graph in
      let print_error buf =
         eprintf "@[<v 3>sort internal error: found too many nodes";
         NodeSet.iter (fun node -> eprintf "@ %a" pp_print_node node) nodes;
         eprintf "@]"
      in
         raise (OmakeException (pos, LazyError print_error))
   else if NodeSet.cardinal marked < NodeTable.cardinal graph then
      let nodes = NodeSet.fold NodeTable.remove graph marked in
      let print_error buf =
         eprintf "@[<v 3>sort internal error: found too many nodes";
         NodeTable.iter (fun node _ -> eprintf "@ %a" pp_print_node node) nodes;
         eprintf "@]"
      in
         raise (OmakeException (pos, LazyError print_error))

let dfs_sort pos graph _ =
   let roots = find_roots pos graph in
   let marked, items =
      NodeTable.fold (fun (marked, items) node _ ->
            dfs_sort_node graph marked items node) (NodeSet.empty, []) roots
   in
      dfs_check pos marked graph;
      List.rev items

(*
 * Check that a list of nodes is in sorted order.
 *)
let check_sort pos graph domain =
   NodeTable.iter (fun node index ->
         let deps = NodeTable.find graph node in
            NodeSet.iter (fun dep ->
                  if NodeTable.find domain dep > index then
                     let print_problem buf =
                        eprintf "@[<hv 3>Nodes are out of order:@ Node %a@ Depends on %a@]" (**)
                           pp_print_node node
                           pp_print_node dep
                     in
                        raise (OmakeException (pos, LazyError print_problem))) deps) domain

(*
 * The main sorting function.
 *)
let sort_aux sorter env venv pos name nodes =
   let pos = string_pos "sort" pos in

   (* Get extra ordering info *)
   let oinfo = venv_get_ordering_info venv name in

   (* Produce a table of the listing order *)
   let domain, _ =
      List.fold_left (fun (domain, i) node ->
            let domain = NodeTable.add domain node i in
            let i = succ i in
               domain, i) (NodeTable.empty, 0) nodes
   in

   (* Build the graph *)
   let graph = build_subgraph env venv pos oinfo domain in
      sorter pos graph domain

(*
 * Top-level functions.
 *)
let check_sort = sort_aux check_sort
let sort = sort_aux dfs_sort

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
